'' I2C Routines  Version 2.2 using parameters in VAR space
'' (c) 2014,15,18 Tracy Allen, EME Systems LLC
'' Terms of MIT liscense, see end of object code
'' Permission is given to use this in any program for the Parallax
'' Propeller processor as long as this copyright notice is included.
'' About this implementation:
''   -- Init(scl,sda) can use any combination of pins for scl,sda.
''   -- both scl and sda need pullups.  The routines manipulate dira[.] only.  Won't work with eeprom on some Parallax development boards.
''   -- Supports clock streching on any rising edge of scl, with timeout parameter and method to change the timeout parameter
''   -- Use 8-bit device addresses, for devSel.  That is the 7-bit address shifted left by 1.
''   -- pin numbers are passed by the Init(scl,sda) method and are stored in VAR space,
''       - subsequent calls to i2c do not pass pin numbers--They use the values stored.
''       - Each object instance declared uses another pair of bytes, but there is only one copy of the object code
''       - every child object must call Init so that it has its own copy of the pins.
''       - Can be used to instantialize multiple buses using different pins, still one copy of the code.
''   -- The hold time defaults to DEFAULT_HOLD_MS, but PUB SetMaxHoldTime(milliseconds) can modify that for each instance.
''   -- 16 bit address can be forced by setting bit 31 (|negx)in the address parameter e.g. ReadByte(devSel, address | i2c#BIGADRS)
''      otherwise it is 8 bit address
''   -- This object does not directly support addresses greater than 16 bits.  That must be handled in application code by patching the device ID.
''   -- negative 1 in address will skip sending address, e.g. eeprom autoincrements, e.g. Readbyte(devsel, i2c#NOADRS)
''   -- The count parameter for strings and numbers can be passed as a negative number. That converts smallEndian<-->largeEndian on the fly.
''   -- Methods for reading and writing numbers (1, 2, 3, 4 ... bytes) can likewise do msbyte<-->lsbyte inversions
''   -- Supports locks. 
''       - For backward compatibility (with my own code base) the lock is set by a separate method.
''       - default is lock disabled by the Init method
''       - a separate method setLock(which) must be called before or AFTER each Init to enable a lock for that instance.
''       - If using locks, be sure all instances that run in separate cogs know their locks before they start running


CON
   ACK      = 0                        ' I2C Acknowledge
   NAK      = 1                        ' I2C No Acknowledge
   DRIVELOW = 1                        ' state of dira[.] is output to drive pin low
   PULLHIGH = 0                        ' state of dira[.] is input to allow pin to be pulled high
   LOW = 0                             ' the output register of both scl and sda will always be low, control is via dira[.] only
   BIGADRS = negx                      ' use word address when device address is or'd with negx ($8000_0000)
   NO_ADDRESS = -1                     ' some commands only devID, no address, or autoincrement, or preset register pointer
   DEFAULT_HOLD_MS = 100               ' wait 100ms for clock hold
   NO_LOCK = -1                        ' the lock variable will run from 1 to 8

   BOOTPIN  = 28                       ' I2C Boot EEPROM SCL Pin
   EEPROM   = $A0                      ' 8-bit device select for I2C EEPROM

OBJ

VAR
  byte scl, sda
  byte errorflag   ' 0 is operation acks, 1 if no ack, mainly for the read operations
  long  maxHoldTicks  ' 8_000_000 is 0.1 second at 80 MHz, for clock stretching
  long lock            ' initialized to zero, no lock unless the SetLock method is called before or after Init

CON ' iniitialization methods
PUB Null
   ' this is not a top level object

PUB Init(_scl,_sda)                               ' pass in both pins, call to set up the buss or to change from one buss to another
  sda := _sda
  scl := _scl
  maxHoldTicks := SetMaxHoldTime(DEFAULT_HOLD_MS)
  lock := 0
  outa[scl] := LOW                                      ' leave a zero in the scl and sda output latches.
  outa[sda] := LOW                                      ' so turning pin to output will drive the pin low.
  dira[scl] := DRIVELOW                                 ' both scl and sda should be pulled high by resistor
  dira[sda] := DRIVELOW                                 ' starting with it a low output to test the pins
  dira[scl] := PULLHIGH                                 ' both scl and sda are pulled high by resistor
  dira[sda] := PULLHIGH
  return  (ina[scl] << 1) | ina[sda]                         ' should return 3 when pullups are present, bus is ready
   
PUB Twiddle                                             ' some devices can use this to get out of a bad state.
  dira[sda] := PULLHIGH
  repeat 9
    dira[scl] := DRIVELOW                             ' Put out up to 9 clock pulses
    dira[scl] := PULLHIGH
  return ina[scl] << 1 & ina[sda]                         ' should return 3 when pullups are present, bus is ready

PUB SetLock(which)     ' a separate routine for backward compatibility
' enter with which = 0 to 7, or NO_LOCK = -1.  
' This method adds one so the variable lock goes from 0 for no-lock and 1 to 8 for the lock number
' The parent object is responsible for checking out and returning locks.
  if which == NO_LOCK
    lock := 0
  else
    lock := which+1   ' add one, so nolock will be 0 or 1 to 8 to use with (if lock then) statement
  
PUB SetMaxHoldTime(milliseconds)                        ' this is for clock stretching, can set independently for each object instance
  maxHoldTicks := clkfreq/1_000 * milliseconds

CON ' low level methods
PUB ReStart                                               ' SDA goes HIGH to LOW with SCL HIGH
  dira[sda] := PULLHIGH
  dira[scl] := PULLHIGH
  if ina[scl] == 0
      Hold
  dira[sda] := DRIVELOW
  dira[scl] := DRIVELOW

PUB Start
  if lock
    repeat while lockset(lock-1)
  Restart
  
PUB Stop                                                ' SDA goes LOW to HIGH with SCL High
  dira[sda] := DRIVELOW
  dira[scl] := PULLHIGH
  if ina[scl] == 0      ' unlikely to need this, will time out
      Hold
  dira[sda] := PULLHIGH
  if lock
    lockclr(lock-1)

PUB Write(bite) : ackbit
'' Write i2c data.  Data byte is output MSB first, SDA data line is valid
'' only while the SCL line is HIGH.  SDA can change only when SCL is low.
'' Data is always 8 bits (+ ACK/NAK).
'' SCL rests in the LOW state during an exchange, sda state is indeterminant
'' returns 0 if slave ACK, 1 if NAK
  ackbit := 0
  bite := !bite << 24
  repeat 8
    dira[sda] := (bite <-= 1)
    dira[scl] := PULLHIGH
    if ina[scl] == 0
      Hold
    dira[scl] := DRIVELOW
  dira[sda] := PULLHIGH         ' allow slave to control sda
  dira[scl] := PULLHIGH         ' this edge latches the ack bit from slave
  if ina[scl] == 0              ' allow slave to hold
    Hold
  ackbit := ina[sda]
  dira[scl] := DRIVELOW         ' sda can change now

PUB Read(ackbit): bite
'' Read in i2c data, Data byte is input MSB first, SDA data line is
'' valid only while the SCL line is HIGH, and changes only when SCL is low
'' SCL left in LOW state, sda state is indeterminant
'' returns byte value
  dira[sda] := PULLHIGH
  repeat 8
    dira[scl] := PULLHIGH
    if ina[scl] == 0
      Hold
    bite := (bite << 1) | ina[sda]
    dira[scl] := DRIVELOW
  dira[sda] := !ackbit
  dira[scl] := PULLHIGH         ' slave latches the ACK or NAK
  if ina[scl] == 0              ' allow slave to hold
    Hold
  dira[scl] := DRIVELOW    ' sda state is indeterminant, device can change sda when scl is low

PRI Hold | ticks
' this implements clock hold, with a timeout
  ticks := cnt + maxHoldTicks
  repeat until (ina[scl]==1) or (cnt - ticks > 0)

PUB Hello(devSel) : ackbit
'' The device will not respond if it is busy, or if it is not present on the i2c buss
'' function call returns 0 if device at devSel responded (ACK, present and ready), otherwise 1 (NAK, not present, or currently busy)
'' note this can return fake ack if the bus is locked low, need also to test for pullups.
   Start
     ackbit := Write(devSel)
   Stop
  
CON ' high level Read methods
PUB ReadPage(devSel, address, dataPtr, byteCount) : ackbit | step1
'' Read in a counted block of i2c data.  Device select code is devSel, 8 bit form with lsb=0
'  Device starting address is "address". Prop address is dataPtr.  Number of bytes is count.
'' address is normally positive 8 bit address; Set high bit (|negx) to force a 16 bit address (EEPROM >65535)
'' For eeproms > 64 kbytes, caller must handle 17th and higher bits ad hoc to merge with devSel
'' Use address=-1 as a flag to skip the device address and go right into the bytes (some devices, MAX1161x, devsel only, selfincrementing memory).
'' The data returned from the device is placed in the Prop byte array pointed to at entry by dataPtr.
'' Normally byteCount is positive and bytes are the same order in Prop as they are in the device.
'' Make count negative to force reverse order, endedness swap.  (benefits for sensors that send msb first)
'' The function call returns 0 if all acks are okay, 1 otherwise, The ACKS/NAKS on read come only from writing devSel and addrReg
'' ----------
  if address <> -1   ' if -1, skip step that sets the device address
    Start                       '
    ackbit := Write(devSel)
    if address & negx             ' bit 31 is set, need to send two byte address
      address &= posx               ' knock it down (not really necessary)
      ackbit |= Write(address >> 8)              ' msb  goin to ac
    ackbit |= Write(address)                    ' lsb or only one address byte
    reStart                        ' Reseltect the device for reading
  else
    Start                          ' Start here if not sending an address
  step1 := 1
  ackbit |= Write(devSel | 1)  ' 8 bit address, 1 in lsb for read operation
  if byteCount < 0                  ' if bytecount is negative, the read will reverse endedness
    -step1
    ||byteCount
    dataPtr += byteCount - 1
  repeat byteCount - 1
    byte[dataPtr] := Read(ACK)   ' same order in both prop and device
    dataPtr += step1           ' step1= +/- 1 for order of bytes moved from device to prop
  byte[dataPtr] := Read(NAK)
  Stop
  ' return 0 if all ACKs okay, 1 if one or more NAKs in the setup
  errorFlag := ackbit

PUB ReadByte(devSel, address)
'' reads one byte from device and returns it to the caller
'' positive value in address sends 8 bit address, value with bit 31 set (|negx) in address sends 16 bit device address
'' -1 in address skips sending any address bytes  (devSel only, followed by repeated start and data read)
   ReadPage(devSel, address, @result, 1)

PUB ReadRegister(devSel)
'' reads one byte from a simple device that needs only device selection.
  Start
  errorFlag := Write(devSel | 1)
  result := Read(NAK)
  Stop

PUB ReadNumber(devSel, address, byteCount)
'' count is length of the number in bytes (1,2,3,4), to be transferred from device to Prop.
'' a long result is returned
'' positive count transfers number byte for byte, negative count swaps the endedness, that is, msb from device goes to lsb in Prop.
'' positive value in address sends 8 bit address, value with bit 31 set (|negx) sends 16 bit address
'' -1 in address skips sending address bytes  (devSel and read at currrent address only)
  ReadPage(devSel, address, @result, byteCount)

PUB getErrorFlag
  return errorFlag    ' in case of error NAK during a read operation


CON ' high level Write methods
PUB WritePage(devSel, address, dataPtr, byteCount) : ackbit | step1
'' Write out a block of i2c data.  Device select code is devSel, 8 bit form.
'  i2c Device starting address or registor is address. Data address is dataPtr.  Number of bytes is count.
'' Special conditions...
''   Use address with high bit set (|negx)to specify 2-byte address (e.g. EEPROM)
''   Use address=-1 to specify null, no address bytes, goes right into the data bytes
''   Use count=0 to skip sending data (i.e. devSel and addrReg only.) (Same as WriteRegister for single byte, but can be two bytes)
''   Use count < 0  to force bytes to be transferred in reverse order, i.e. first byte out becomes last byte in the target.  Endedness swap.
'' For eeproms > 64 kbytes, caller must handle 17th and higher bits ad hoc to merge with devSel
'' Return zero if all acks are okay, 1 otherwise
'' Page boundaries are  the responsibility of the caller!!!!!   See object eeprom for examples that handle page boundaries

'' ----------
  Start
  step1 := 1
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  if address <> -1         ' if -1 skip address altogether, direct to sending bytes
    if address & negx             ' high bit set signals 2 byte address (eeprom)
      address &= posx        ' knock it down, not really necessary
      ackbit |= Write(address >> 8)   ' msb
    ackbit |= Write(address)           ' lsb, or the one and only address byte
  if byteCount<0                                            ' edit1, consolidated +/- with step1 variable
    -step1
    ||byteCount
    dataptr += byteCount - 1
  repeat byteCount
     ackbit |= Write(byte[dataPtr])
     dataPtr += step1                                  ' step1= +/- 1 for order of bytes moved from prop to device
  Stop
  ' return 0 if all ACKs okay, 1 if one or more NAKs

PUB WriteRegister(devSel, mybyte) : ackbit
'' Writes one byte to a device register.
'' This might be a very simple device, or the register might be a pointer or config register.
'' Or it might be eeprom where the address autoincrements
'' Returns 0 if all ACKs are okay, 1 if one or more NAKs
  Start
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  ackbit |= Write(mybyte)
  Stop


PUB WriteByte(devSel, address, myByte) : ackbit
'' writes one byte to device.  This is faster than WriteNumber for a single byte
'' set the bit31 (|negx)in order to send 16 bit address
'' use address = -1 to skep sending address altogether (e.g. sequential memory access)
'' if address is not needed, us WriteRegister instead of WriteByte
'' function call returns 0 if all ACKs are okay, 1 if one or more NAKs
  Start
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  if address & negx
    address &= posx        ' knock it down, not really necessary
    ackbit |= Write(address>>8)
  ackbit |= Write(address)    ' lsbyte or only byte
  ackbit |= Write(myByte)
  Stop

PUB WriteNumber(devSel, address, number, byteCount)
'' count is length of the number in bytes, 1--4, to be transferred from Prop to device.
'' count as a posibive number transfers number byte for byte, negative count swaps the endedness, that is, msb from prop goes to lsb in device.
'' set bit31 of address (|negx) to send 16 bit address
'' -1 in address skips sending any address bytes  (sends devSel only then the number for autoincrementing device)
'' function call returns 0 if all ACKs are okay, 1 if one or more NAKs
   return  WritePage(devSel, address, @number, byteCount)


DAT
   ' none     ' 

CON

{{
*==============================================================================================================================*
*                                                   TERMS OF USE: MIT License                                                  *
*------------------------------------------------------------------------------------------------------------------------------*
*Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    *
*files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    *
*modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software*
*is furnished to do so, subject to the following conditions:                                                                   *
*                                                                                                                              *
*The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.*
*                                                                                                                              *
*THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          *
*WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         *
*COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   *
*ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         *
*==============================================================================================================================*
}}

