''=============================================================================
'' @file     sht31_1bus.spin   version 1.0
'' @target   Propeller with Sensirion SHT3x
'' @author   Thomas Tracy Allen, Manpreet Gill, EME Systems
'' Copyright (c) 2016 EME Systems LLC
'' See end of file for terms of use.
'' 
'' This object is Spin only, no extra COGs.  Uses subsidiary I2C object.
''
''=============================================================================
CON
                          '
' The SHT31 has two possible device IDs, as follows
  DEV_ID44 = $44 << 1
  DEV_ID45 = $45 << 1

' this object supports single shot mode with clock hold (stretch)
' At three different repeatability/conversion times
  HOLD_HIGH = $2c06        ' 0.1%RH, 12 milliseconds conversion time
  HOLD_MEDIUM = $2c0d      ' 0.15% RH, 4 milliseconds
  HOLD_LOW = $2c10         ' 0.25% RH, 2 milliseconds

' This object supports the following control actions
  HEATER_ON = $306d       ' heater control
  HEATER_OFF = $3066      ' heater control
  SOFT_RESET = $30A2      ' reloads calibration data, not sure if it resets modes
  
' this object does NOT support
' -- time delay or ack polling for single shot mode (supports clock stretching only)
' -- periodic (autonomous) modes, including the ALERT function and the Read Status command.


VAR
  long device_id, mode, degC, rh, maxholdms

OBJ
  uarts  : "fullDuplexSerial4pnf"   'Debug terminal for communication
  i2c    : "i2c_1bus"

PUB Null
  ' this is not a top level object

PUB Init(_scl, _sda, _device_ID, _mode)
  result := i2c.init(_scl, _sda)
  result := (result << 1) + i2c.hello(_device_ID)   ' return value should be 6 for pullups high + successful ack.
  device_ID := _device_ID
  mode := _mode
  
PUB SetLock(which)
  i2c.SetLock (which)

PUB AcquireData | temp1, temp2
  ' acquire temperature and humidity with clock hold (stretch)
  ' sends device ID and two byte command
  ' then SHT31 holds and returns 6 bytes temperature and RH data with crc's.
  ' values returned to local variables temp1 and temp2
  ' ignore the checksum
  temp2 := 0
  i2c.ReadPage(device_ID, mode | i2c#BIGADRS, @temp1, -6)  ' 
  ' -6 puts bytes in array in swapped order.   temp1:(rhcrc, rh0, rh1, tccrc) temp2:(tc0, tc1,0,0)
  degC :=  temp2 * 100 ** 11468975 - 4500    ' xxx.xx    raw * 175/65535 - 45. returns degC to 100ths
  rh := (temp1<<8>>16) ** 65536000           ' xxx.x     raw * 100 / 65535, shifts clear out the checksums, data to tenths.

PUB Temperature
  ' results from the lasst AcquireData command
  return degC                   ' units of xxx.xx degC

PUB Humidity
  ' results from the lasst AcquireData command
  return rh                     ' units of xx.x %RH

PUB SetMaxHoldTime(ms)
  i2c.SetMaxHoldTime(ms)

PUB Control(action)
' these are the control commands to turn on or off the heater or to do soft reset.
  i2c.Start              'start i2c
  i2c.write(device_ID) ' devSel is 8 bit, zero in lsb
  i2c.write(action.byte[1])  ' preface for the control command, will be $30
  i2c.write(action.byte[0])
  i2c.stop 

PUB Reset
  return i2c.Twiddle   ' return should equal 3 for buss released high.

PUB Pause(ms)
  waitcnt(clkfreq/1000*ms+cnt)
  
PUB Hello   ' check for presence
  return i2c.hello(device_ID)
  

DAT
    ' no DAT section
    ' 
CON 
{AcquireDataSteps | temp1, temp2
  leaving this in as comment to illustrate an alternative AcquireData method using small steps with generic I2C driver
  i2c.Start              'start i2c
  i2c.write(device_ID) ' devSel is 8 bit, zero in lsb
  i2c.write($2c)  ' this is the command to stretch the clock, single shot
  i2c.write($0D)
  ' i2c.stop  '  don't need this in clock stretch mode, can go right to repeated start.
  ' for polled modes, insert the stop plus a delay here, or poll until ack.
  i2c.start     ' repeated start
  i2c.Write(device_ID | 1)
  temp1 := I2C.read(i2c#ACK)   'reading the first byte
  temp1 := (temp1 << 8) + I2C.READ(i2c#ACK) 'first two bytes are temperature readings
  I2C.READ(i2c#ACK)             '3rd byte is a checksum
  temp2 := I2C.READ(i2c#ACK)      ' next two bytes contain the humidity readings
  temp2 := (temp2 << 8) + I2C.READ(i2c#ACK)
  I2C.READ(i2c#ACK)             '3rd byte is a checksum
  i2c.stop               ' stop i2c
  degC :=  temp1 * 100 ** 11468975 - 4500       ' xxx.xx    raw * 175/65535 - 45
  rh := temp2 ** 65561000   ' xxx.x    raw * 100 / 65535
}

CON
{{
┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}

